import time
import tarfile
import os
from .utils import check_dir
from .backends import ssh


cache_dir = 'build/__cache__'


def set_server(ip, target_location, port=22):
    ssh.set_server(ip, target_location, port)


def set_user(username, password):
    ssh.set_user(username, password)


def pack(source_dir, cache_dir):
    source_files = [x for x in os.listdir(source_dir)]
    assert 'main.py' in source_files
    check_dir(cache_dir)
    target_fp = f'{cache_dir}/deploy_pack.tar.gz'
    with tarfile.open(target_fp, "w:gz") as tar:
        for root, _, files in os.walk(source_dir):
            for fname in files:
                pathfile = f'{root}/{fname}'
                arcpath = f'{root.split(source_dir)[-1]}/{fname}'
                tar.add(pathfile, arcname=arcpath)
    print(f'packing {target_fp}')
    return target_fp


def deploy(src='src', run='main.py', conda_env=None):
    global cache_dir
    deploy_pack = pack(src, cache_dir)
    fsize = os.path.getsize(deploy_pack)
    f_mb = fsize/float(1024)/float(1024)
    print(f'package size: {f_mb*1024:.6f}KB ({f_mb:.4f}MB)')
    if f_mb > 10:
        print('src too large for ssh.')
        raise NotImplementedError
    try:
        assert ssh.server_ip is not None
    except Exception:
        print('server_ip not specified, use labvision.io.ssh.set_server() to init remote server.')
        return
    try:
        assert ssh.server_location is not None
    except Exception:
        print('target_location not specified, use labvision.io.ssh.set_server() to init remote server.')
        return
    try:
        assert ssh.user is not None or ssh.password is not None
    except Exception:
        print('user or password not specified, use labvision.io.ssh.set_user() to init remote server.')
        return
    print('transfering deploy package to remote ssh ..')
    ssh.upload(deploy_pack, f'{ssh.server_location}/deploy_pack.tar.gz')
    ssh.exec_command(f'cd {ssh.server_location}')
    ssh.exec_command(f'tar -zxvf deploy_pack.tar.gz')
    if conda_env:
        ssh.exec_command(f'source activate {conda_env};nohup python {run} >log.d 2>&1 &')
    else:
        ssh.exec_command(f'screen -s labvision-remote;nohup python {run} >log.d 2>&1 &')
    print(f'\t[remote] successfully deployed, running in background now. (pid={os.getpid()})')
    print(f"\t[remote] you can use 'cd {ssh.server_location};cat log.d' to see the log.")
    print(f'removing cached deploy package: {deploy_pack}')
    os.remove(deploy_pack)


def run(command):
    ssh.exec_command(command)


def close(latency=1):
    print(f'\t[remote] ssh close in {latency} secs ..')
    time.sleep(latency)
    ssh.close()



import random
import torch
import labvision
from .utils import MissingKeyException
from .controller import AutoControl


def __update__(config, default):
    for k, v in default.items():
        if k not in config:
            config[k] = v
        elif type(v) is dict and type(config[k]) is dict:
            config[k] = __update__(config[k], v)
    return config


def default_config():
    random.seed()
    config = {
        'seed': random.randint(0, 65536),
        'model': None,
        'default_optimizer': torch.optim.SGD,
        'optimizer': {
            'lr': 1e-3,
            'weight_decay': 5e-4,
        },
        'datasets': {
            'dataset_class': None,
            'args': {
                'transform': labvision.transforms.resize_rdcrop_flip(224, (224, 224)),
            },
            'trainset': None,
            'testset': None,
            'valset': None,
            'batch_size': 64,
            'num_workers': 2,
        },
        'frozen_status': None,
        'log_file_path': 'rec.log',
        'build_dir': 'build',
    }
    return config


def check_config(config):
    if config['model'] is None:
        raise MissingKeyException('model', 'A model must be specified in config.')
    datasets = config['datasets']
    if datasets['dataset_class'] is None:
        if datasets['trainset'] is None:
            raise MissingKeyException('datasets.trainset',
                                      'You must specify both trainset and testset, or set use_simple_def=True')
        if datasets['testset'] is None:
            raise MissingKeyException('datasets.testset',
                                      'You must specify both trainset and testset, or set use_simple_def=True')


def compile(config=None, read_frozen_status=False, **kwargs):
    if config is None:
        config = kwargs  # kwargs mode

    # fill default values
    config = __update__(config, default_config())
    check_config(config)

    # init optimizer
    model = config['model']
    optimizer = config['optimizer']
    if type(optimizer) is dict:
        default_optim = config['default_optimizer']
        optimizer = default_optim(model.parameters(), **optimizer)
        config['optimizer'] = optimizer

    # init datasets
    datasets = config['datasets']
    if datasets['dataset_class'] is not None:
        trainset = datasets['dataset_class'](train=True, **datasets['args'])
        testset = datasets['dataset_class'](train=False, **datasets['args'])
        datasets['trainset'] = trainset
        datasets['testset'] = testset
    num_workers = datasets['num_workers']
    num_workers = min(num_workers, labvision.io.backends.device.cpu_num_workers_limit)
    datasets['num_workers'] = num_workers

    # load frozen status
    if not read_frozen_status:
        config['frozen_status'] = None

    return AutoControl(config)


def resume(fp):
    config = torch.load(fp)
    ac = AutoControl(config)
    ac.__log__(f'continue from: {fp}', pure_log=True)
    return ac



from . import models
from . import datasets
from . import transforms
from . import io
from . import auto
from . import visualize

__all__ = ['models', 'datasets', 'transforms', 'io', 'auto', 'visualize']



def acc_topk(logits, y, ks=(1, 3)):
    """
        Args:
            logits:
            y:
            ks: top-k accs to be evaluated.
    """
    metrics = {f'acc@top{k}': 0 for k in ks}
    for k in ks:
        metrics[f'acc@top{k}'] += (logits.topk(max((1, k)), 1, True, True)[1] == y.view(-1, 1)).sum().float().item()/logits.shape[0]
    return metrics



from . import utils
from .basic import empty_transform
from .basic import resize_centercrop_flip
from .basic import resize_rdcrop_flip
from . import casnet_cvpr_2018

__all__ = ['utils', 'empty_transform', 'resize_centercrop_flip', 'resize_rdcrop_flip',
           'casnet_cvpr_2018']



import paramiko


server_ip = None
server_port = None
server_location = ''
port = None
user = None
password = None
ssh = None
command_location = '~'


def set_server(ip, target_location, port=22):
    global server_ip, server_location, server_port
    server_ip = ip
    server_location = target_location
    server_port = port


def set_username(_str):
    global user
    user = _str


def set_password(_str):
    global password
    password = _str


def set_user(username, pwd):
    set_username(username)
    set_password(pwd)


def connect():
    global server_ip, server_port, user, password
    # SSH远程连接
    ssh = paramiko.SSHClient()  # 创建sshclient
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())  # 指定当对方主机没有本机公钥的情况时应该怎么办，AutoAddPolicy表示自动在对方主机保存下本机的秘钥
    ssh.connect(server_ip, server_port, user, password)
    print(f'\t[remote] connected to {user}@{server_ip}:{server_port}')
    return ssh


def close():
    global ssh
    if ssh is not None:
        print('\t[remote] ssh closed.\n')
        ssh.close()


def cd(location):
    global ssh, command_location
    if ssh is None:
        ssh = connect()
    print(f'\t[remote] server:{command_location}$ cd {location}')
    _, stdout, stderr = ssh.exec_command(f'cd {location};pwd',)
    command_location = stdout.readlines()[0].replace('\n', '')
    return stdout, stderr


def exec_command(command):
    global ssh, command_location
    if ssh is None:
        ssh = connect()
    if command.startswith('cd'):
        return cd(command[3:])
    print(f'\t[remote] server:{command_location}$ {command}')
    _, stdout, stderr = ssh.exec_command(f'. /etc/profile;. .bashrc;cd {command_location};{command}')
    out = stdout.readlines()
    for line in out:
        print(f'\t[remote] server:{command_location}$ {line}', end='')
    err = stderr.readlines()
    for line in err:
        print(f'\t[remote] server:{command_location}$ {line}', end='')
    return out, err


def upload(local_file, remote_file):
    global ssh
    if ssh is None:
        ssh = connect()
    exec_command(f'mkdir {server_location}')
    sftp = ssh.open_sftp()
    sftp.put(local_file, remote_file)


if __name__ == "__main__":
    set_server('172.20.46.235', '/home/sh/Desktop/remote_deploy')
    set_user('sh', 'sh')
    exec_command('ls')
    close()
